package it.polito.bigdata.spark.exercise2;
import org.apache.spark.api.java.*;

import scala.Tuple2;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.spark.SparkConf;

public class SparkDriver {

    public static void main(String[] args) {
        
		Logger.getLogger("org").setLevel(Level.OFF);
		Logger.getLogger("akka").setLevel(Level.OFF);

        String neObjectData;
        String observationData;
        String outputPath1;
        String outputPath2;

        neObjectData = "exam_ex2_data/NEObject.txt";
        observationData = "exam_ex2_data/Observations.txt";
        
        outputPath1 = "outPart1/";
        outputPath2 = "outPart2/";

        // Initialize SparkContext
        SparkConf conf = new SparkConf().setAppName("Exam 20240205_Spark").setMaster("local");
		// Create a Spark Context object
		JavaSparkContext sc = new JavaSparkContext(conf);

        // Load RDDs data from files
        JavaRDD<String> neObjectRDD = sc.textFile(neObjectData);
        JavaRDD<String> observationsRDD = sc.textFile(observationData);

        // ########################################
        // PART 1
        // ########################################

        // Find the average Dimension of all NEObjects
        JavaRDD<Tuple2<Double, Integer>> neObjectMeasures = neObjectRDD.map(line -> {
            // Assuming NEObject.txt format is: 
            // NEOID,Dimension,MaterialStrength,alreadyFallen
            String[] fields = line.split(",");
            Double dimension = Double.parseDouble(fields[1]);
            return new Tuple2<>(dimension, 1);
        });

        // Compute the average dimension using a reduce action
        Tuple2<Double, Integer> sumAndCount = neObjectMeasures.reduce((v1, v2) -> {
            Double dimSum = v1._1() + v2._1();
            Integer countSum = v1._2() + v2._2();
            return new Tuple2<Double, Integer>(dimSum, countSum);
        });

        Double dimensionAvg = sumAndCount._1() / sumAndCount._2();

        // Analyze NEObject by filtering by dimension and alreadyFallen fields.
        // The resulting RDD contains the Most Relevant NEOIDs
        // Key = NEOID
        // Value = +1
        JavaPairRDD<String, Integer> mostRelevantNeoId = neObjectRDD.filter(line -> {
            // NEOID,Dimension,MaterialStrength,alreadyFallen

            String[] fields = line.split(",");
            Double dimension = Double.parseDouble(fields[1]);
            Boolean isFallen = Boolean.parseBoolean(fields[3]);

            return dimension > dimensionAvg && !isFallen;
        }).mapToPair(line -> {
            String[] fields = line.split(",");
            String neoid = fields[0];

            return new Tuple2<String,Integer>(neoid,1);
        }).cache();

        // Keep only Observations made in 2023 or later
        // and map data into a pairRDD with
        // key = NEOID
        // value = ObservatoryID
        JavaPairRDD<String, String> observationsFiltered = observationsRDD.filter(line -> {
            // NEOID,ObservatoryID,ObsDateTime,EclipticLat,EclipticLon,EstimatedDistance
            String[] fields = line.split(",");
            String date = fields[2];
            Integer year = Integer.parseInt(date.split("-")[0]);

            return year >= 2023;
        }).mapToPair(line -> {
            String[] fields = line.split(",");
            String neoid = fields[0];
            String observatoryId = fields[1];
            return new Tuple2<String, String>(neoid, observatoryId);
        }).cache();

        // Join the two PairRDDs, obtaining all the observations in 2023 or later 
        // associated with the Most Relevant NEOIDs
        // key = NEOID
        // value = (ObservatoryID, +1)
        // and map it to
        // key = NEOID
        // value = +1
        // and use a reduceByKey to count the number of observations per 
        // Most RelevantNEOID
        JavaPairRDD<String, Integer> observationsPerNeoId = observationsFiltered.join(mostRelevantNeoId)
                                                    .mapValues(v -> 1)
                                                    .reduceByKey((v1, v2) -> v1 + v2);

        // swap key and value, sort it in descending order and swap again
        JavaPairRDD<String, Integer> res1 = observationsPerNeoId.mapToPair(t -> t.swap())
                                                .sortByKey(false)
                                                .mapToPair(t -> t.swap());

        res1.saveAsTextFile(outputPath1);


        // ########################################
        // PART 2
        // ########################################
        // To compute the most relevant NEOIDs observed by less than 10 distinct observatories, 
        // we can reuse the previously computed RDDs
        // observationsFiltered: contains the observatories per each NEOID
        // mostRelevantNeoId: contains the Most Relevant NEOIDs

        // First, compute the list of NEOIDs with >= 10 distinct observatories starting from 2023
        // by performing a distinct over observationsFiltered RDD to compute the distinct observatories per NEOID
        // and by counting those
        // key = NEOID
        // value = count
        // Then, we keep only the NEOIDs with count >= 10, which are the ones that should be discarded
        JavaPairRDD<String, String> distinctObservatoriesPerNeoId = observationsFiltered
                                                                        .distinct().cache();

        JavaPairRDD<String, Integer> neoIdsWithManyObservatories = distinctObservatoriesPerNeoId
                                                        .mapValues(v -> 1)
                                                        .reduceByKey((v1, v2) -> v1 + v2)
                                                        .filter(v -> v._2() >= 10);
        
        // From the complete list of Most Relevant NEOIDs, remove NEOIDs in the previously computed RDD
        // key = NEOID
        // value = +1
        JavaPairRDD<String, Integer> neoIdsOfInterest = mostRelevantNeoId
                                    .subtractByKey(neoIdsWithManyObservatories);

        // Start from observations made in 2023 by distinct observatories for each NEOID,
        // and use an outerJoin to keep only the entries which satisfies the two conditions
        // - most relevant NEOID
        // - number of distinct observatories <= 10 (0 included!)
        // missing values (i.e., no observations) are filled with "None"
        JavaPairRDD<String, String> res2 = distinctObservatoriesPerNeoId
                                                        .rightOuterJoin(neoIdsOfInterest)
                                                        .mapValues(t -> t._1().orElse("None"));

        res2.saveAsTextFile(outputPath2);


        // Close SparkContext
        sc.close();
    }
}
