package it.polito.bigdata.spark.exercise2;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.spark.sql.Dataset;
import org.apache.spark.sql.Row;
import org.apache.spark.sql.SparkSession;
import org.apache.spark.sql.types.DataTypes;


public class SparkDriver {

    public static void main(String[] args) {
        
		Logger.getLogger("org").setLevel(Level.OFF);
		Logger.getLogger("akka").setLevel(Level.OFF);

        String neObjectData;
        String observationData;
        String outputPath1;
        String outputPath2;


        neObjectData = "exam_ex2_data/NEObject.txt";
        observationData = "exam_ex2_data/Observations.txt";
        
        outputPath1 = "outPart1/";
        outputPath2 = "outPart2/";

		// Create a Spark Session object and set the name of the application
		SparkSession ss = SparkSession.builder().appName("Exam20240205 Spark")
									.master("local").getOrCreate();

		// Define the dataframes associated with the input files
		Dataset<Row> neosDF = ss.read().format("csv")
				.option("header", true).option("inferSchema", true)
				.load(neObjectData);


		// Register the temporary table NEOs 
		neosDF.createOrReplaceTempView("NEOs");


		Dataset<Row> observationsDF = ss.read().format("csv")
				.option("header", true).option("inferSchema", true)
				.load(observationData);


		// Register the temporary table Observations 
		observationsDF.createOrReplaceTempView("Observations");

        // ########################################
        // PART 1
        // ########################################

        // Compute the average dimension considering all NEObjects
        Dataset<Row> avgDimensionDF = ss.sql("SELECT AVG(Dimension) as AvgDim "
                                           + "FROM NEOs");
    
        // Register the temporary table avgDimension
        avgDimensionDF.createOrReplaceTempView("avgDimension");

        // Analyze the content of NEOs by filtering by dimension and alreadyFallen fields.
        // The resulting RDD contains the Most Relevant NEOIDs
        Dataset<Row> mostRelNEOsDF = ss.sql("SELECT NEOID "
                                           + "FROM NEOs, avgDimension "
                                           + "WHERE Dimension>AvgDim "
                                           + "  AND alreadyFallen=false")
                                           .cache();

        // Register the temporary table mostRelNEOs
        mostRelNEOsDF.createOrReplaceTempView("mostRelNEOs");
        
		// Register a function that given a month returns the year
		ss.udf().register("EXTRACTYEAR", 
				(String month) -> Integer.parseInt(month.split("-")[0]),
				DataTypes.IntegerType);		
        
        // Keep only Observations made in 2023 or later
        // and select NEOID and ObservatoryID (the second field is used in the second part of this application).
        Dataset<Row> observationsFrom2023DF = ss.sql("SELECT NEOID, ObservatoryID "
                                           + "FROM Observations "
                                           + "WHERE EXTRACTYEAR(ObsDateTime)=2023")
                                           .cache();

        // Register the temporary table observationsFrom2023
        observationsFrom2023DF.createOrReplaceTempView("observationsFrom2023");

        // Join observationsFrom2023 with mostRelNEOs, obtaining all the observations in 2023 or later 
        // associated with the Most Relevant NEOIDs.
        // Then, count the number of observations starting from 2023 for each Most Relevant NEOID.
        // Finally, sort data by number of observations (descending)
        Dataset<Row> observationsPerNeoIdDF = ss.sql("SELECT mostRelNEOs.NEOID, COUNT(*) "
                                           + "FROM observationsFrom2023, mostRelNEOs "
                                           + "WHERE mostRelNEOs.NEOID=observationsFrom2023.NEOID "
                                           + "GROUP BY mostRelNEOs.NEOID "
                                           + "ORDER BY COUNT(*) DESC");


        observationsPerNeoIdDF.write()
            .format("csv")
            .option("header", false)
            .save(outputPath1);


        // ########################################
        // PART 2
        // ########################################
        // To compute the most relevant NEOIDs observed by less than 10 distinct observatories, 
        // we can reuse the previously computed tables
        // observationsFrom2023: contains the observatories per each NEOID
        // mostRelNEOs: contains the Most Relevant NEOIDs

        // Perform a distinct over observationsFrom2023 to find the distinct observatories per NEOID,
        // starting from 2023
        Dataset<Row> distinctObservatoriesPerNeoIdDF = ss.sql("SELECT DISTINCT * "
                                                + "FROM observationsFrom2023")
                                                .cache();
        
                                                // Register the temporary table distinctObservatoriesPerNeoId
        distinctObservatoriesPerNeoIdDF.createOrReplaceTempView("distinctObservatoriesPerNeoId");



        // First, compute the list of NEOIDs with >= 10 distinct observatories starting from 2023 (inner query).
        // Then, keep only the most relevant NEOs that are not in that set (i.e, discard those with count >= 10.
        // The kept most relevant NEOs are those of interest.
        Dataset<Row> neoIdsOfInterestDF = ss.sql("SELECT NEOID "
                                                + "FROM mostRelNEOs "
                                                + "WHERE NEOID NOT IN "
                                                + "      (SELECT NEOID "
                                                + "       FROM distinctObservatoriesPerNeoId "
                                                + "       GROUP BY NEOID "
                                                + "       HAVING COUNT(*)>=10)");
        
        // Register the temporary table neoIdsOfInterest
        neoIdsOfInterestDF.createOrReplaceTempView("neoIdsOfInterest");


	
        // Register a function that given a null value returns "NONE"
		ss.udf().register("FUNCNONE", 
				(String ObsId) -> { if (ObsId==null) return "NONE";
                                    else return ObsId; },
				DataTypes.StringType);		

        // Start from the observations made starting from 2023 by distinct observatories for each NEOID,
        // and use an outerJoin to keep only the entries which satisfies the two conditions
        // - most relevant NEOID
        // - number of distinct observatories <= 10 (0 included!)
        // missing values (i.e., no observations) are filled with "None"


        Dataset<Row> res2DF = ss.sql("SELECT neoIdsOfInterest.NEOID, FUNCNONE(ObservatoryID) "
                                   + "FROM distinctObservatoriesPerNeoId RIGHT OUTER JOIN neoIdsOfInterest "
                                   + "ON distinctObservatoriesPerNeoId.NEOID=neoIdsOfInterest.NEOID");

        res2DF.write()
            .format("csv")
            .option("header", false)
            .save(outputPath2);

        ss.stop();
    }
}
